/* ------------------------- linec.c ---------------------------------------- */

/* Draw lines on SVGA. Can be cery fast when usage of low level assembly
 * support is requested. It handles VGA page crossing.
 *
 * It originated from the DJgpp package but completely re-written since, the
 * only surviving clue being the WriteMode specification in the color field.
 * Actually this is not used either.
 *
 * These functions are pure C:
 *
 * GrMoveToC (x, y)
 * GrDrawToC (x, y, color)
 * GrLineC (x0, y0, x1, y1, color)
 * GrPolygonC (npoints, points, color)
 *
 * If you only use the low-level asm version then this module is redundant.
*/

#include <stdlib.h>
#include <string.h>

#include "grh.h"


#define FAST_PLOT(va) \
	if (!_GrWriteMode)			\
		*va = (uchar)c;			\
	else if (_GrWriteMode & GrXOR)		\
		*va ^= c;			\
	else /* if (_GrWriteMode & GrXOR) */	\
		*va |= c

#define LINE_PLOT(va,vb) \
	if ((int)(MY_OFF(va)^vb) < 0) {				\
		if (MY_OFF(va) > 0xc000U)			\
			--Page, _GrSetPage (Page, Page);	\
		else if (MY_OFF(va) < 0x4000U)			\
			++Page, _GrSetPage (Page, Page);	\
	}							\
	FAST_PLOT (va)

static uint	x1 = 0, y1 = 0, xydone = 0;
static uchar	far *va = VGA_PAGE;


extern uchar far * far FASTCALL
_GrAddrC (uint x, uint y)
{
	ulong	xy;
	ushort	page;

	page = (ushort)((xy = x + _ActiveBase + y*(long)_GrSizeX) >> 16);
#if 0
	if (page > 0x0f)
		return VGA_PAGE;
#endif
	if (page != _WritePage)
		_GrSetPage (page, page);

	return (VGA_PAGE + (ushort)xy);
}

#ifdef USE_OLD_ILOOP
static int far
_InnerLoopOld (uint iva, int dy, int dx, int dvx, int dvy, int c)
{
	register int	count, err;
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _WritePage;

	MY_OFF(va) = iva;
	err = dx / 2;
	for (count = dx; --count >= 0;) {
		vb = MY_OFF(va);
		MY_OFF(va) += dvx;
		err += dy;
		if (err > dx) {
			err -= dx;
			MY_OFF(va) += dvy;
		}
		LINE_PLOT (va,vb);
	}
	return (MY_OFF(va));
}

#else

static int far
_InnerLoop (uint iva, int dy, int dx, int dvx, int dvy, int c)
{
	register long	err;		/* conceptualy */
	register int	count;
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _WritePage;

	/* dx > dy  never =!!!	*/
	/* dx > 0 		*/
	/* dy >= 0		*/

	/* set:			*/
	/* dy = (dy*0x10000)/dx	*/
	/* dx = 0x10000		*/

	MY_OFF(va) = iva;
	count = dx;

#ifdef USE_DIVTAB
	dy = dy * divtab[dx];
#else
	dy = (int)((0x00010000L*dy)/dx);
#endif

	err = 0x00008000L;
	for (; --count >= 0;) {
		vb = MY_OFF(va);
		MY_OFF(va) += dvx;
		if ((err += (ulong)(uint)dy) & 0xffff0000L) {
			err &= 0x0000ffffL;
			MY_OFF(va) += dvy;
		}
		LINE_PLOT (va,vb);
	}
	return (MY_OFF(va));
}
#endif

static int far
_SimpleLoop (uint vi, register int count, register int sv, int c)
{
	uchar		far *va = VGA_PAGE;
	uint		vb;
	int		Page = _WritePage;
	

	MY_OFF(va) = vi;
	while (--count >= 0) {
		vb = MY_OFF(va);
		MY_OFF(va) += sv;
		LINE_PLOT (va,vb);
	}
	return (MY_OFF(va));
}

extern void far
GrMoveToC (uint x, uint y)
{
#if 0
	if (x >= _GrSizeX)
		x = _GrSizeX-1;
	if (y >= _GrSizeY)
		y = _GrSizeY-1;
#endif
	x1 = x;
	y1 = y;
	xydone = 0;
}

extern void far
GrDrawToC (uint x2, uint y2, uint c)
{
	register int	dx, dy, svx, svy;
	int		t;

#if 0
	if (x2 >= _GrSizeX)
		x2 = _GrSizeX-1;
	if (y2 >= _GrSizeY)
		y2 = _GrSizeY-1;
#endif
	t = 0;
	if (!xydone) {
		t = 1;
		xydone = 1;
		MY_OFF(va) = (ushort)_GrAddrC (x1, y1);
		FAST_PLOT(va);
	}

	svx = 1;
	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		svx = -svx;
	}

	svy = _GrSizeX;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		svy = -svy;
	}

	if (dx > dy) {
		if (0 == dy) {
			int	Page;

			Page = _WritePage;

			if (_GrWriteMode)
				MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx,
						c);
			else if (svx >= 0) {
				ushort	room;

				room = 0xffffU - MY_OFF(va);
				if (dx <= room) {
					memset (va+1, c, dx);
					MY_OFF(va) += dx;
				} else {
					if (room > 0)
						memset (va+1, c, room);
					++Page, _GrSetPage (Page, Page);
					MY_OFF(va) = dx - room - 1;
					memset (VGA_PAGE, c, MY_OFF(va)+1);
				}
			} else {
				if (MY_OFF(va) >= dx) {
					MY_OFF(va) -= dx;
					memset (va, c, dx);
				} else {
					memset (VGA_PAGE, c, MY_OFF(va));
					--Page, _GrSetPage (Page, Page);
					dx -= MY_OFF(va);
					MY_OFF(va) = (ushort)(0x10000UL-dx);
					memset (va, c, dx);
				}
			}
		} else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dy, dx, svx, svy,
					c);
		++GrStats[dx+t];
	} else if (dx < dy) {
		if (0 == dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dy, svy, c);
		else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dx, dy, svy, svx,
					c);
		++GrStats[dy+t];
	} else /* if (dx == dy) */ {
		if (dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx+svy, c);
		++GrStats[dx+t];
	}

	x1 = x2;
	y1 = y2;
}

extern void far
GrLineC (uint xa, uint ya, uint xb, uint yb, uint c)
{
	if (xa != x1 || ya != y1 || !xydone)
		GrMoveToC (xa, ya);
#if 0
	if ((c&0xff00) != _GrWriteMode)
		GrSetWriteMode (c);
#endif
	GrDrawToC (xb, yb, c);
}

extern void far
GrClearC (uint x, uint y, uint sx, uint sy, uint color)
{
	long	room;
	int	Page;

	if (!sx || !sy)
		return;

	if (!_GrWriteMode) {
		MY_OFF(va) = (ushort)_GrAddrC (x, y);
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
		Page = _WritePage;
		for (; sy--; ++y) {
			if (room <= 0) {
				room += 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			if ((long)sx > room) {
				memset (va, color, (uint)room);
				++Page, _GrSetPage (Page, Page);
				memset (VGA_PAGE, color, sx-(uint)room);
				MY_OFF(va) = _GrSizeX - (uint)room;
				room = 0x10000UL - MY_OFF(va);
			} else {
				memset (va, color, sx);
				MY_OFF(va) += _GrSizeX;
				room -= _GrSizeX;
			}
		}
		GrMoveToC (x+sx-1, y-1);
	} else {
		for (sx += x-1; sy--; ++y) {
			GrMoveToC (x,  y);
			GrDrawToC (sx, y, color);
		}
	}
}

#define muldiv(x,y,z)	((int)(((long)(x) * (long)(y)) / (long)(z)))

extern void far
GrPolygonC (int npoints, ushort *points, uint color)
{
	uint	y;			/* current y */
	int	fwdx, fwdy, bwdx, bwdy;	/* current edge tops */
	int	fdx, fdy, bdx, bdy;	/* interpolation */
	int	fx, bx;			/* segment endpoints */
	ushort	*lastpoint;		/* end of 'points' */
	ushort	*fwd, *bwd;		/* current point */
	ushort	*nfwd, *nbwd;		/* next point */
	long	room;
	uint	n;
	int	Page;

	if (npoints < 2)
		return;

	if (2 == npoints) {
		GrMoveToC (points[1], points[2]);
		GrDrawToC (points[3], points[4], color);
		return;
	}

/* locate the top vertex (use fwd as temp).
*/
	lastpoint = points + 2*(npoints - 1);
	y = (fwd = points)[1];
	for (bwd = points+2; bwd <= lastpoint; bwd += 2) {
		if (y > bwd[1])
			y = (fwd = bwd)[1];
	}

/* run through the edges, top to bottom.
*/
	nfwd = nbwd = fwd;

	if (!_GrWriteMode) {
		xydone = 0;		/* don't bother keeping currency */
		MY_OFF(va) = (ushort)_GrAddrC (0, y);
		Page = _WritePage;
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
	}

	for (;;) {
		if (y == nfwd[1]) {
			do {
				fwd = nfwd;
				if (fwd == lastpoint)
					nfwd = points;
				else
					nfwd = fwd + 2;
			} while (y == nfwd[1]);
			if ((fdy = nfwd[1] - fwd[1]) < 0)
				break;
			fdx = nfwd[0] - (fwdx = fwd[0]);
			fwdy = 0;
		}

		if (y == nbwd[1]) {
			do {
				bwd = nbwd;
				if (bwd == points)
					nbwd = lastpoint;
				else
					nbwd = bwd - 2;
			} while (y == nbwd[1]);
			if ((bdy = nbwd[1] - bwd[1]) < 0)
				break;
			bdx = nbwd[0] - (bwdx = bwd[0]);
			bwdy = 0;
		}

		fx = fwdx + muldiv (fwdy, fdx, fdy);
		bx = bwdx + muldiv (bwdy, bdx, bdy);
		if (!_GrWriteMode) {
			if (fx > bx) {
				n  = fx;
				fx = bx;
				bx = n;
			}
			n = bx - fx + 1;
			if ((long)fx >= room)
				++Page, _GrSetPage (Page, Page);
			MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
			room = 0x10000UL - MY_OFF(va);
			if ((long)n > room) {
				memset (va, color, (uint)room);
				n -= (uint)room;
				MY_OFF(va) = 0;
				room = 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			memset (va, color, n);
			n += _GrSizeX - bx - 1;
			MY_OFF(va) += n;
			room -= n;
		} else {
			GrMoveToC (fx, y);
			GrDrawToC (bx, y, color);
		}
		++y;
		++fwdy;
		++bwdy;
	}

/* do bottom tip/edge.
*/
	fx = fwdx + fdx;
	bx = bwdx + bdx;
	if (!_GrWriteMode) {
		if (fx > bx) {
			n  = fx;
			fx = bx;
			bx = n;
		}
		n = bx - fx + 1;
		if ((long)fx >= room)
			++Page, _GrSetPage (Page, Page);
		MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
		room = 0x10000UL - MY_OFF(va);
		if ((long)n > room) {
			memset (va, color, (uint)room);
			n -= (uint)room;
			MY_OFF(va) = 0;
			++Page, _GrSetPage (Page, Page);
		}
		memset (va, color, n);
		GrMoveToC (bx, y);
	} else {
		GrMoveToC (fx, y);
		GrDrawToC (bx, y, color);
	}
}

#undef muldiv
